-- Invariants used by both semiRealTimeBankersQueue.agda and
-- bankersQueue.agda. Note the following two things:
-- 1) bankersQueue.agda requires trickier analysis than
--    semiRealTimeBankersQueue.agda.
-- 2) Some proofs below would have been less complicated if m had been
--    0 (as it is in semiRealTimeBankersQueue.agda), or at least any
--    constant (2 would have sufficed for both bankersQueue.agda and
--    semiRealTimeBankersQueue.agda). The price for generality wasn't
--    too high, though, and hard-coding constants does not feel nice.

-- # import eq
-- # import prelude
-- # import nat
-- # import vec
-- # import order
-- # import cumSum

bqInvariant : (m, ℓ, Δ : ℕ) -> Vec ℓ ℕ -> Set
bqInvariant m ℓ Δ Cs =
  (i : ℕ) -> (le : i ≤ ℓ) -> cumSum i le Cs ≤ min ((m + 1) * i) Δ

bqRaiseΔ (m, ℓ, Δ, Δ' : ℕ) (Cs : Vec ℓ ℕ)
  : Δ ≤ Δ' -> bqInvariant m ℓ Δ Cs -> bqInvariant m ℓ Δ' Cs
bqRaiseΔ leΔ inv i le = inv i le `transitive` minMonotone reflexive leΔ

bqNil : (m, Δ : ℕ) -> bqInvariant m 0 Δ []
bqNil _ _ _ leZ = leZ

bqSingleton : (m, Δ : ℕ) -> bqInvariant m 1 Δ [0]
bqSingleton _ _ _ leZ       = leZ
bqSingleton _ _ _ (leS leZ) = leZ

bqTail (m, ℓ, Δ, C : ℕ) (Cs : Vec ℓ ℕ)
  :  bqInvariant m (ℓ + 1) Δ (C :: Cs)
  -> bqInvariant m ℓ (Δ ∸ (m + 1)) (Cs ⊖ (m + 1 ∸ C))
bqTail {m} {ℓ} {Δ} {C} {Cs} inv i le =
  -- cumSum i le (Cs ⊖ (m + 1 ∸ C)) ≡ (C + cumSum i le Cs) ∸ (m + 1)
  liftEq (congBin (+) (sym prop) refl
            `trans`
          prop_cumSum_sub (m + 1) (leS le) (C :: Cs)
         )
    `transitive`
  -- (C + cumSum i le Cs) ∸ (m + 1) ≤
  -- (min ((m + 1) * (i + 1)) Δ) ∸ (m + 1)
  subtractMonotone {m + 1} (inv (i + 1) (leS le))
    `transitive`
  -- (min ((m + 1) * (i + 1)) Δ) ∸ (m + 1) ≡
  -- min ((m + 1) * i) (Δ ∸ (m + 1))
  liftEq (homomorphicMin (\_ _ -> subtractMonotone {m + 1})
                         ((m + 1) * (i + 1)) Δ
            `trans`
          congBin min (prop' (m + 1) i) refl
         )
  where
  prop : C ∸ (m + 1) ≡ 0
  prop = rewrite-plugin(sym $ mulRightOne {m},
           subtractLarger (inv 1 (leS leZ) `transitive` minLeft)
         )

  prop' : (x, y : ℕ) -> x * (y + 1) ∸ x ≡ x * y
  prop' x y =
    congBin (∸) (mulRightSuc {x} {y}) refl
      `trans`
    subtractPlusInverses' {x} {y * x}
      `trans`
    mulCommutative

bqInit (m, Δ, ℓ : ℕ) (Cs : Vec (ℓ + 1) ℕ)
  :  bqInvariant m (ℓ + 1) Δ Cs
  -> bqInvariant m ℓ Δ (init Cs)
bqInit {m} {Δ} {ℓ} {Cs} inv i le =
  liftEq (cong sum $ sym $ prop_take_init i le Cs)
    `transitive`
  inv i (plusLeftIncreasing {1} le)

bqLast (m, C, ℓ, Δ : ℕ) (Cs : Vec ℓ ℕ)
  :  sum Cs + C ≤ (m + 1) * (ℓ + 1)
  -> bqInvariant m ℓ Δ Cs
  -> bqInvariant m (ℓ + 1) (Δ + C) (Cs ++ [C])
bqLast {_} {_} {ℓ} _ inv i iLe with i ≤?> ℓ =>
  _ {m} {C} {ℓ} {Δ} {Cs} _ inv i iLE₁ (left iLE₂) =
    liftEq (cong sum $ cong (\le -> take i le (Cs ++ [C]))
                            (leUnique iLE₁ $ plusRightIncreasing {1} iLE₂)
                         `trans`
                       prop_take_append₁ i iLE₂ Cs [C]
           )
      `transitive`
    inv i iLE₂
      `transitive`
    minMonotone reflexive plusIncreasingRight
  _ {m} {C} {ℓ} {Δ} {Cs} cscLE inv i iLE (right iGT) =
    liftEq (cong (\iLE -> cumSum (witness iLE) (witnessOK iLE) (Cs ++ [C]))
                 prop₂
              `trans`
            cong sum (prop_take_all (Cs ++ [C]))
              `trans`
            prop_sum_append Cs [C]
            )
      `transitive`
    leMinDistrib₂ rewrite-plugin(prop₁, cscLE)
                  (plusMon (liftEq (cong sum $ sym $ prop_take_all Cs)
                              `transitive`
                            inv ℓ reflexive
                              `transitive`
                            minRight
                           )
                           reflexive
                  )
    where
    prop₁ : i ≡ ℓ + 1
    prop₁ = antisymmetric iLE iGT

    prop₂ : Exists {ℕ} {\i -> i ≤ ℓ + 1} i iLE ≡
            Exists {ℕ} {\i -> i ≤ ℓ + 1} (ℓ + 1) reflexive
    prop₂ = existsCong prop₁ (leUnique _ _)

bqSub (m, ℓ, Δ : ℕ) (Cs : Vec ℓ ℕ)
  : (n : ℕ) -> bqInvariant m ℓ (Δ + n) Cs -> bqInvariant m ℓ Δ (Cs ⊖ n)
bqSub {m} {ℓ} {Δ} {Cs} n inv _ le =
  liftEq (prop_cumSum_sub n le Cs)
    `transitive`
  subtractMonotone (inv _ le)
    `transitive`
  liftEq (homomorphicMin (\_ _ -> subtractMonotone) _ _)
    `transitive`
  minMonotone (subtractDecreasing reflexive)
              (liftEq subtractPlusInverses)

bqAppend (ℓ₁, ℓ₂, Δ₁, Δ₂, m : ℕ) (Cs₁ : Vec ℓ₁ ℕ) (Cs₂ : Vec ℓ₂ ℕ)
  :  bqInvariant m ℓ₁ Δ₁ Cs₁
  -> bqInvariant m ℓ₂ Δ₂ Cs₂
  -> bqInvariant m (ℓ₁ + ℓ₂) (Δ₁ + Δ₂) (Cs₁ ++ Cs₂)
bqAppend {ℓ₁} inv₁ inv₂ i le with i ≤?≥ ℓ₁ =>
  _ {ℓ₁} {ℓ₂} {Δ₁} {Δ₂} {_} {Cs₁} {Cs₂} inv₁ inv₂ i le₁ (left  le₂) =
    liftEq (cong sum $ cong (\le -> take i le (Cs₁ ++ Cs₂))
                            (leUnique le₁ $ plusRightIncreasing le₂)
                         `trans`
                       prop_take_append₁ i le₂ Cs₁ Cs₂
           )
      `transitive`
    inv₁ i le₂
      `transitive`
    minMonotone reflexive plusIncreasingRight
  _ {ℓ₁} {ℓ₂} {Δ₁} {Δ₂} {m} {Cs₁} {Cs₂} inv₁ inv₂ i le₁ (right le₂) =
    liftEq (cong sum (prop_take_append₂ i le₂ le₁ Cs₁ Cs₂)
              `trans`
            helper _
              `trans`
            prop_sum_append _ _
           )
      `transitive`
    plusMon rewrite-plugin(sym $ prop_take_all Cs₁, inv₁ ℓ₁ reflexive)
            (inv₂ (minus le₂) (prop_minus_plus le₁ le₂))
    -- ≤ min ((m + 1) * ℓ₁) Δ₁ + min ((m + 1) * minus le₂) Δ₂
      `transitive`
    bihomomorphicMin' plusMon' _ _ _ _
    -- ≤ min ((m + 1) * ℓ₁ + (m + 1) * minus le₂) (Δ₁ + Δ₂)
      `transitive`
    minMonotone (liftEq (sym (mulDistribLeftAdd {m + 1})
                           `trans`
                         cong ((*) (m + 1)) (prop_plus_minus le₂)
                        )
                )
                reflexive
    -- ≤ min ((m + 1) * i) (Δ₁ + Δ₂)
    where
    helper (ℓ₁, ℓ₂ : ℕ) (xs : Vec ℓ₁ ℕ)
      : (eq : ℓ₁ ≡ ℓ₂) -> sum (subst (Vec' ℕ) eq xs) ≡ sum xs
    helper refl = refl

bqReplicate (m, C, ℓ : ℕ)
  : C ≤ m + 1 -> bqInvariant m ℓ (C * ℓ) (replicate {ℓ} C)
bqReplicate                 cLE 0       leZ       = leZ
bqReplicate {m} {C} {suc ℓ} cLE (suc i) (leS iLE) =
  -- C + sum (take i iLE (replicate C))
  plusMon reflexive (bqReplicate cLE i iLE)
  -- C + (min ((m + 1) * i) (C * ℓ))
    `transitive`
  liftEq (homomorphicMin (\_ _ -> plusRightMon {C}) _ _)
  -- min (C + (m + 1) * i) (C + C * ℓ)
    `transitive`
  minMonotone (plusMon cLE reflexive
                 `transitive`
               liftEq (sym $ mulRightSuc' {m + 1} {i})
              )
              (liftEq (congBin (+) refl mulCommutative
                         `trans`
                       mulCommutative {ℓ + 1} {C}
                      )
              )
  -- min ((m + 1) * (i + 1)) (C * suc ℓ)

bqZero (m, ℓ : ℕ) (Cs : Vec ℓ ℕ)
  : bqInvariant m ℓ 0 Cs -> Cs ≡ replicate 0
bqZero {_} {_} {[]}      inv = refl
bqZero {_} {_} {C :: Cs} inv with inv _ (leS leZ) =>
  _ {m} {suc ℓ} inv leZ =
    congBin (::) refl $ bqZero {m} {ℓ} (helper {_} {m} inv)
    where
    helper (ℓ, m : ℕ) (Cs : Vec ℓ ℕ)
      : bqInvariant m (ℓ + 1) 0 (0 :: Cs) -> bqInvariant m ℓ 0 Cs
    helper inv' _ le = inv' _ (leS le)

bqSuc (m, ℓ : ℕ) (Cs : Vec ℓ ℕ)
  :  bqInvariant m ℓ 0 Cs
  -> bqInvariant m ℓ ℓ (map suc Cs)
bqSuc {m} inv = rewrite-plugin(bqZero {m} inv, helper _ _)
  where
  helper : (ℓ, m : ℕ) -> bqInvariant m ℓ ℓ (map suc $ replicate 0)
  helper _       _ _       leZ      = leZ
  helper (suc ℓ) _ (suc i) (leS le) = leS $
    helper ℓ _ i le
      `transitive`
    minMonotone (plusMon reflexive $ mulMon reflexive sucIncreasing)
                reflexive

bqRot (m, ℓ : ℕ) (Cs : Vec ℓ ℕ)
  :  bqInvariant m ℓ 0 Cs
  -> bqInvariant m (2 * ℓ + 1) ℓ (map suc Cs ++ replicate {ℓ + 1} 0)
bqRot {m} {ℓ} {Cs} inv =
  bqAppend {ℓ} {ℓ + 1} {ℓ} {0} {m} (bqSuc inv) (bqReplicate {m} leZ)

bqAppendRev (m, ℓ : ℕ) (Cs : Vec (ℓ + 1) ℕ)
  :  bqInvariant (m + 2) (ℓ + 1) 0 Cs
  -> bqInvariant (m + 2) (2 * ℓ + 3) (3 * ℓ + 5)
                 ((map suc (init Cs) ++ [ℓ + 3]) ++ replicate {ℓ + 2} 1)
bqAppendRev {m} {ℓ} {Cs} inv =
  subst (\ℓ' -> bqInvariant (m + 2) (2 * ℓ + 3) ℓ'
                  ((map suc (init Cs) ++ [ℓ + 3]) ++ replicate {ℓ + 2} 1))
        (cong ((+) 5) $ sym addAssociative) $
  bqAppend {ℓ + 1} {ℓ + 2} {2 * ℓ + 3} {ℓ + 2} {m + 2}
           (bqLast {m + 2} {ℓ + 3} prop $
              bqSuc {m + 2} $ bqInit {m + 2} {0} inv)
           (bqReplicate {m + 2} (leS leZ))
  where
  prop₂ : sum (map suc (init (0 :: replicate 0))) ≡ ℓ
  prop₂ = cong sum (cong (map suc) (prop_init_replicate ℓ)
                      `trans`
                    prop_map_replicate suc
                   )
            `trans`
          prop_sum_replicate ℓ 1
            `trans`
          mulRightSuc

  prop : sum (map suc (init Cs)) + (ℓ + 3) ≤ (m + 3) * (ℓ + 1)
  prop = rewrite-plugin(bqZero {m + 2} inv,
           rewrite-plugin(prop₂,
             plusMon (reflexive {ℓ + 3}) plusIncreasingRight
           )
         )

bqAppendRev' (m, ℓ : ℕ) (Cs : Vec (ℓ + 1) ℕ)
  :  bqInvariant (m + 2) (ℓ + 1) 0 Cs
  -> bqInvariant (m + 2) (2 * ℓ + 3) (2 * (2 * ℓ + 3))
                 (map suc (init Cs) ++ (ℓ + 3) :: replicate {ℓ + 2} 1)
bqAppendRev' {m} {ℓ} {Cs} inv =
  helper {m + 2} {_} {2 * ℓ + 3} {2 * (2 * ℓ + 3)} _ $
  subst (bqInvariant (m + 2) (2 * ℓ + 3) (2 * (2 * ℓ + 3)))
        (sym $ prop_append_assoc (map suc (init Cs)) [ℓ + 3]
                                 (replicate {ℓ + 2} 1)) $
  bqRaiseΔ {m + 2} (prop _) $ bqAppendRev {m} inv
  where
  prop : (n : ℕ) -> 3 * n + 5 ≤ 2 * (2 * n + 3)
  prop n = plusMon (plusIncreasingRight {1} {5})
                   (plusMon (plusIncreasingRight {n} {n}) reflexive)

  helper (m', ℓ₁, ℓ₂, Δ : ℕ) (Cs' : Vec ℓ₁ ℕ)
    :  (eq : ℓ₁ ≡ ℓ₂)
    -> bqInvariant m' ℓ₂ Δ (subst (Vec' ℕ) eq Cs')
    -> bqInvariant m' ℓ₁ Δ Cs'
  helper refl = id
