------------------------------------------------------------------------
-- Substitutions
------------------------------------------------------------------------

module Substitution where

open import Kind
open import TypeSystem
open import Erasure as E

open import Data.Function using (id)
open import Data.Nat using (_+_)

------------------------------------------------------------------------
-- General substitution machinery

-- Uses idea from "Type-Preserving Renaming and Substitution" by Conor
-- McBride to enable defining both term and variable substitutions
-- without having to write several functions which traverse terms.

-- Definition of what a substitution is.

module SubstDef (_•_ : TmLike) -- The substitution replaces variables
                               -- with this kind of "term".
                where

  infix  10 _⇒_
  infixl 20 _▷_

  -- Substitutions of type Γ ⇒ Δ, when applied, take things with
  -- variables in Γ and give things with variables in Δ. (This concept
  -- is sometimes written with the arrow in the other direction.)

  data _⇒_ {k : Kind} : Ctxt k -> Ctxt k -> Set where
    ∅   : forall {Δ}     -> ε ⇒ Δ
    _▷_ : forall {Γ Δ τ} -> Γ ⇒ Δ -> Δ • τ -> Γ ▻ τ ⇒ Δ

  -- The type of functions applying substitutions to things.

  Applier : TmLike -> TmLike -> Set
  Applier _•₁_ _•₂_ =    forall {k} {Γ Δ : Ctxt k} {τ}
                      -> Γ •₁ τ -> Γ ⇒ Δ -> Δ •₂ τ

  -- You get lots of substitution machinery defined for you (see
  -- below) if you define a SubstKit, ...

  record SubstKit : Set where
    field
      vr     :  forall {k} {Γ : Ctxt k} {τ}   -> Γ ∋ τ -> Γ • τ
      weaken :  forall {k} {Γ : Ctxt k} {σ τ} -> Γ • τ -> Γ ▻ σ • τ
      tm     :  forall {k} {Γ : Ctxt k} {τ}   -> Γ • τ -> Γ ⊢ τ
      wait   :  forall {Γ τ n₁} C n₂
             -> Γ • C [ Thunk n₁ τ ] -> Γ • C [ Thunk (n₂ + n₁) τ ]

  -- ...and more if you define a SubstKit⁺.

  record SubstKit⁺ : Set where
    field
      kit  : SubstKit
      _/•_ : Applier _•_ _•_

open SubstDef using (SubstKit; SubstKit⁺)

-- Various substitution-related functions.

module Subst {_•_ : TmLike}
             (kit : SubstKit _•_)
             where

  open SubstDef _•_ public
  open SubstKit kit

  infix  70 _↑
  infixl 48 _/⊢_ _/∋_

  -- Weakening of substitutions.

  wk⇒ : forall {k} {Γ Δ : Ctxt k} {σ} -> Γ ⇒ Δ -> Γ ⇒ Δ ▻ σ
  wk⇒ ∅       = ∅
  wk⇒ (ρ ▷ t) = wk⇒ ρ ▷ weaken t

  -- Lifting of substitutions.

  _↑ : forall {k} {Γ Δ : Ctxt k} {σ} -> Γ ⇒ Δ -> Γ ▻ σ ⇒ Δ ▻ σ
  ρ ↑ = wk⇒ ρ ▷ vr vz

  -- Identity.

  idˢ : forall {k} {Γ : Ctxt k} -> Γ ⇒ Γ
  idˢ {Γ = ε}     = ∅
  idˢ {Γ = Γ ▻ τ} = idˢ ↑

  -- Weakening.

  wk : forall {k} {Γ : Ctxt k} {σ} -> Γ ⇒ Γ ▻ σ
  wk = wk⇒ idˢ

  -- Substitution of a single variable.

  sub : forall {k} {Γ : Ctxt k} {τ} -> Γ • τ -> Γ ▻ τ ⇒ Γ
  sub t = idˢ ▷ t

  -- Substitution application for variables.

  _/∋_ : Applier _∋_ _•_
  vz           /∋ (ρ ▷ t) = t
  vs x         /∋ (ρ ▷ t) = x /∋ ρ
  gwait∋ C m x /∋ ρ       = wait C m (x /∋ ρ)

  -- Substitution application for terms.

  _/⊢_ : Applier _⊢_ _⊢_
  var x        /⊢ ρ = tm (x /∋ ρ)
  λ t          /⊢ ρ = λ (t /⊢ ρ ↑)
  λ² t         /⊢ ρ = λ² (t /⊢ ρ ↑ ↑)
  t₁ · t₂      /⊢ ρ = (t₁ /⊢ ρ) · (t₂ /⊢ ρ)
  fix t        /⊢ ρ = fix (t /⊢ ρ ↑)
  z            /⊢ ρ = z
  s            /⊢ ρ = s
  natrec t₁ t₂ /⊢ ρ = natrec (t₁ /⊢ ρ) (t₂ /⊢ ρ ↑ ↑)
  (t₁ , t₂)    /⊢ ρ = (t₁ /⊢ ρ) , (t₂ /⊢ ρ)
  uncurry t    /⊢ ρ = uncurry (t /⊢ ρ ↑ ↑)
  return t     /⊢ ρ = return (t /⊢ ρ)
  (t₁ >>= t₂)  /⊢ ρ = t₁ /⊢ ρ >>= t₂ /⊢ ρ
  (√ t)        /⊢ ρ = √ t /⊢ ρ
  pay n t      /⊢ ρ = pay n (t /⊢ ρ)
  gpay C n t   /⊢ ρ = gpay C n (t /⊢ ρ)
  gwait C n t  /⊢ ρ = gwait C n (t /⊢ ρ)

-- Subst⁺ defines composition. This function uses a stronger
-- substitution kit, which is sometimes defined using the Subst module
-- (see the examples below); this is the reason for placing
-- composition in a separate module.

module Subst⁺ {_•_  : TmLike}
              (kit⁺ : SubstKit⁺ _•_)
              where

  open SubstDef.SubstKit⁺ _•_ kit⁺
  open Subst kit public

  infixl 70 _∘ˢ_

  _∘ˢ_ : forall {k} {Γ Δ X : Ctxt k} -> Γ ⇒ Δ -> Δ ⇒ X -> Γ ⇒ X
  ∅        ∘ˢ ρ₂ = ∅
  (ρ₁ ▷ t) ∘ˢ ρ₂ = ρ₁ ∘ˢ ρ₂ ▷ (t /• ρ₂)

-- Erasing terms in substitutions.

module Erasure (• : TmLike) (⌜_⌝• : TmEraser •) where

  open SubstDef •

  ⌜_⌝⇒ : forall {Γ Δ} -> Γ ⇒ Δ -> ⌜ Γ ⌝▻ ⇒ ⌜ Δ ⌝▻
  ⌜ ∅ ⌝⇒     = ∅
  ⌜ ρ ▷ t ⌝⇒ = ⌜ ρ ⌝⇒ ▷ ⌜ t ⌝•

------------------------------------------------------------------------
-- Instantiations of the general structures above

-- Variables.

varKit : SubstKit _∋_
varKit = record
  { vr     = id
  ; weaken = vs
  ; tm     = var
  ; wait   = gwait∋
  }

varKit⁺ : SubstKit⁺ _∋_
varKit⁺ = record { kit  = varKit
                 ; _/•_ = Subst._/∋_ varKit
                 }

module VarSubst where
  open Subst⁺ varKit⁺ public
  open Erasure _∋_ ⌜_⌝∋ public

-- Terms.

-- Note that term substitutions are not used by the formalisation.
-- They are included to illustrate the utility of the machinery
-- defined above, and to show that it is possible to define full term
-- substitutions.

tmKit : SubstKit _⊢_
tmKit = record
  { vr     = var
  ; weaken = \t -> t /⊢ wk
  ; tm     = id
  ; wait   = gwait
  }
  where open VarSubst

tmKit⁺ : SubstKit⁺ _⊢_
tmKit⁺ = record { kit  = tmKit
                ; _/•_ = Subst._/⊢_ tmKit
                }

module TmSubst where
  open Subst⁺ tmKit⁺ public
  open Erasure _⊢_ ⌜_⌝ public
